/*
 * HttpD.c
 *
 *      Author: Trax
 */

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <avr/interrupt.h>
#include "Net/uip/uip.h"
#include "Net/uip/uip_TcpAppHub.h"
#include "Net/uip/uip_UdpAppHub.h"
#include <avr/eeprom.h>
#include "shell.h"
#include "Ports.h"
#include "Options.h"
#include "Screens.h"
#include "Settings.h"
#include "Hardware/Lcd.h"

#define SHELL_PROMPT PSTR("NetIO> ")

void shell_start(struct telnetd_state* s)
{
  shell_output(s, 1, PSTR("NetIO command shell"));
  shell_output(s, 0, SHELL_PROMPT);
}

uint8_t g_ID = 0;
uint8_t g_OP = 0;
void shell_queue(void)
{
	switch(g_OP)
	{
	case 1:
		AddOption(0);
		break;
	case 2:
		UpOption(g_ID);
		break;
	case 3:
		RemoveOption(g_ID);
		break;
	case 4:
		DownOption(g_ID);
		break;
	case 5:
		AddScreen();
		break;
	case 6:
		UpScreen(g_ID);
		break;
	case 7:
		RemoveScreen(g_ID);
		break;
	case 8:
		DownScreen(g_ID);
		break;
	case 9:
		Save(g_ID);
	}

	g_OP = 0;
	g_ID = 0;
}

int GetComboVal(const char* pTest, const prog_char* pOptions);
uint8_t GetPort(const char *pID);
char* GetComboOpt(int pOpt, const prog_char* pOptions, char* pBuff);


static const prog_char g_Help[] = "Help";
static const prog_char g_HelpText[] = "Read the source code file: shell.c";

static const prog_char g_Setup[] = "Setup";

static const prog_char g_DHCP[] = "DHCP %S";
static const prog_char g_On[] = "On";
static const prog_char g_Off[] = "Off";
static const prog_char g_Renew[] = "Renew";

static const prog_char g_IP[] = "IP";
static const prog_char g_Netmask[] = "Netmask";
static const prog_char g_Gateway[] = "Gateway";

static const prog_char g_Config[] = "Config";

static const prog_char g_SetIP[] = "SetIP";
static const prog_char g_SetNM[] = "SetNM";
static const prog_char g_SetGW[] = "SetGW";

static const prog_char g_Reboot[] = "Reboot";
static const prog_char g_Quit[] = "Quit";

static const prog_char g_Port[] = "Port %S: %s";

static const prog_char g_Lcd1[] = "Lcd %S: %s";
static const prog_char g_Lcd2[] = "Lcd %S: %i";
static const prog_char g_Lcd3[] = "LCD Pin %s: %s";

static const prog_char g_Opt[] = "Option %i %S: %s";
static const prog_char g_OptN[] = "Option %i %S: %i";

static const prog_char g_Val[] = "%S %i: %s";
static const prog_char g_ValN[] = "%S %i: %i";

const prog_char g_Label[] = "Label";
const prog_char g_IO[] = "IO";

const prog_char g_Support[] = "Support";
const prog_char g_Lines[] = "Lines";
const prog_char g_LCDPin[] = "LCDPin";

const prog_char g_AddOP[] = "AddOP";
const prog_char g_UpOP[] = "UpOP";
const prog_char g_RemOP[] = "RemOP";
const prog_char g_DownOP[] = "DownOP";

const prog_char g_Input[] = "Input";
const prog_char g_InVal[] = "InVal";
const prog_char g_InAux[] = "InAux";
const prog_char g_Output[] = "Output";
const prog_char g_OutVal[] = "OutVal";

const prog_char g_Value[] = "Value";

const prog_char g_AddPG[] = "AddPG";
const prog_char g_UpPG[] = "UpPG";
const prog_char g_RemPG[] = "RemPG";
const prog_char g_DownPG[] = "DownPG";

const prog_char g_Page[] = "Page";
const prog_char g_Write[] = "Write";

const prog_char g_Action[] = "Action";
const prog_char g_Auto[] = "Auto";

const prog_char g_Save[] = "Save";

const prog_char g_Hi[] = "Hi";
const prog_char g_Lo[] = "Lo";
const prog_char g_NULL[] = "";

static const prog_char g_Unknown[] = "Unknown command: %s";



extern const prog_char g_OptsIOA[];
extern const prog_char g_OptsIO[];
extern const prog_char g_OptsIOP[];
extern const prog_char g_OptsMEM[];
extern const prog_char g_OptsALL[];

extern const prog_char g_OptsLCD[];

extern const prog_char g_OptsBitV[];
extern const prog_char g_OptsBinV[];
extern const prog_char g_OptsAdcV[];

static const prog_char g_IPs[] = "%d.%d.%d.%d";
static const prog_char g_IPp[] = "%S %d.%d.%d.%d";

uint8_t ReadIP(char* buffer, uip_ipaddr_t *__a)
{
    int Ip[4];
	if(sscanf_P(buffer,g_IPs,&Ip[0],&Ip[1],&Ip[2],&Ip[3]) == 4)
	{
		uip_ipaddr(*__a, Ip[0], Ip[1], Ip[2], Ip[3]);
		return 1;
	}
	return 0;
}

extern char g_Buffer[128];
void shell_input(char *pBuff, struct telnetd_state* s)
{
	uip_ipaddr_t IpAddr;

	// Zerlege den befehl
	char* BuffPtr = g_Buffer;
	uint8_t ID = 0xFF;
	uint8_t Update = 0;
	while(*pBuff)
	{
		if(*pBuff < '0' || *pBuff > '9')
		{
			if(*pBuff == '=' || *pBuff == ' ')
			{
				Update = 1;
				pBuff++;
				break;
			}
			*BuffPtr++ = *pBuff++;
		}
		else if(ID == 0xFF)
			ID = atoi(pBuff);
		else
			pBuff++;
	}
	*BuffPtr = 0;

	//dekodiere die %xx browser schreibweise
	int j,k,l;
	for(j=k=0;pBuff[j] != 0;j++)
	{
		if(pBuff[j] == '+')
			pBuff[k] = ' ';
		else if(pBuff[j] == '%')
		{
			sscanf(pBuff+j+1,"%2x",&l);
			pBuff[k] = (char)l;
			j+=2;
		}
		else if(k!=j)
			pBuff[k] = pBuff[j];
		k++;
	}
	pBuff[k] = 0;

	// Identifiziere den befehl und fhre ihn aus
	if(strcmp_P(g_Buffer,g_Help) == 0){
		shell_output(s, 1, g_HelpText);
	}

	else if(strcmp_P(g_Buffer,g_Setup) == 0)
	{
		if(eeprom_read_byte((uint8_t*)EE_DC))
			shell_output(s, 1, g_DHCP, g_On);
		else
			shell_output(s, 1, g_DHCP, g_Off);

		eeprom_read_IP4(EE_IP, &IpAddr);
		shell_output(s, 1, g_IPp, g_IP, uip_ipaddr1(IpAddr), uip_ipaddr2(IpAddr), uip_ipaddr3(IpAddr), uip_ipaddr4(IpAddr));

		eeprom_read_IP4(EE_NM, &IpAddr);
		shell_output(s, 1, g_IPp, g_Netmask, uip_ipaddr1(IpAddr), uip_ipaddr2(IpAddr), uip_ipaddr3(IpAddr), uip_ipaddr4(IpAddr));

		eeprom_read_IP4(EE_GW, &IpAddr);
		shell_output(s, 1, g_IPp, g_Gateway, uip_ipaddr1(IpAddr), uip_ipaddr2(IpAddr), uip_ipaddr3(IpAddr), uip_ipaddr4(IpAddr));
	}
	else if(strcmp_P(g_Buffer,g_Config) == 0)
	{
		uip_gethostaddr(IpAddr);
		shell_output(s, 1, g_IPp, g_IP, uip_ipaddr1(IpAddr), uip_ipaddr2(IpAddr), uip_ipaddr3(IpAddr), uip_ipaddr4(IpAddr));

		uip_getnetmask(IpAddr);
		shell_output(s, 1, g_IPp, g_Netmask, uip_ipaddr1(IpAddr), uip_ipaddr2(IpAddr), uip_ipaddr3(IpAddr), uip_ipaddr4(IpAddr));

		uip_getdraddr(IpAddr);
		shell_output(s, 1, g_IPp, g_Gateway, uip_ipaddr1(IpAddr), uip_ipaddr2(IpAddr), uip_ipaddr3(IpAddr), uip_ipaddr4(IpAddr));
	}
	else if(strncmp_P(g_Buffer,g_DHCP,4) == 0)
	{
		if(strcmp_P(pBuff,g_On) == 0){
			eeprom_write_byte((uint8_t*)EE_DC, 1);
			shell_output(s, 1, g_DHCP, g_On);
		}else if(strcmp_P(pBuff,g_Off) == 0){
			eeprom_write_byte((uint8_t*)EE_DC, 0);
			shell_output(s, 1, g_DHCP, g_Off);
		}else if(strcmp_P(pBuff,g_Renew) == 0){
			shell_output(s, 1, g_DHCP, g_Renew);
			dhcpc_request();
		}else{
			if(eeprom_read_byte((uint8_t*)EE_DC))
				shell_output(s, 1, g_DHCP, g_On);
			else
				shell_output(s, 1, g_DHCP, g_Off);
		}
	}
	else if(strcmp_P(g_Buffer,g_SetIP) == 0)
	{
		if(ReadIP(pBuff,&IpAddr))
		{
			uip_sethostaddr(IpAddr);
			eeprom_write_IP4(EE_IP, &IpAddr);
			shell_output(s, 1, g_IPp, g_IP, uip_ipaddr1(IpAddr), uip_ipaddr2(IpAddr), uip_ipaddr3(IpAddr), uip_ipaddr4(IpAddr));
		}
	}
	else if(strcmp_P(g_Buffer,g_SetNM) == 0)
	{
		if(ReadIP(pBuff,&IpAddr))
		{
			uip_setnetmask(IpAddr);
			eeprom_write_IP4(EE_NM, &IpAddr);
			shell_output(s, 1, g_IPp, g_Netmask, uip_ipaddr1(IpAddr), uip_ipaddr2(IpAddr), uip_ipaddr3(IpAddr), uip_ipaddr4(IpAddr));
		}
	}
	else if(strcmp_P(g_Buffer,g_SetGW) == 0)
	{
		if(ReadIP(pBuff,&IpAddr))
		{
			uip_setdraddr(IpAddr);
			eeprom_write_IP4(EE_GW, &IpAddr);
			shell_output(s, 1, g_IPp, g_Gateway, uip_ipaddr1(IpAddr), uip_ipaddr2(IpAddr), uip_ipaddr3(IpAddr), uip_ipaddr4(IpAddr));
		}
	}
	else if(strcmp_P(g_Buffer,g_Reboot) == 0)
	{
		cli(); // Disable all interrupts
		do { for(;;){} } while(0); // Wait for the watchdog
	}
	else if(strcmp_P(g_Buffer,g_Quit) == 0)
	{
		if(s){
			shell_quit(s);
			return;
		}
	}

	else if(strcmp_P(g_Buffer,g_Label) == 0){
		if(Update){
			SetPortLabel(ID, pBuff);
		}else{
			shell_output(s, 1, g_Port, g_Label, GetPortLabel(ID,1));
		}
	}
	else if(strcmp_P(g_Buffer,g_IO) == 0){
		if(Update){
			SetPortIO(ID, GetComboVal(pBuff, g_OptsALL));
		}else{
			shell_output(s, 1, g_Port, g_IO, GetComboOpt(GetPortIO(ID),g_OptsALL,g_Buffer));
		}
	}

	else if(strcmp_P(g_Buffer,g_Support) == 0){
		if(Update){
			SetLCDSupport(GetComboVal(pBuff, g_OptsLCD)+1);
		}else{
			shell_output(s, 1, g_Lcd1, g_Support, GetComboOpt(GetLCDSupport()-1,g_OptsLCD,g_Buffer));
		}
	}
	else if(strcmp_P(g_Buffer,g_Lines) == 0){
		if(Update){
			SetLCDLines(atoi(pBuff));
		}else{
			shell_output(s, 1, g_Lcd2, g_Lines, GetLCDLines());
		}
	}
	else if(strcmp_P(g_Buffer,g_LCDPin) == 0){
		if(Update){
			SetLCDPin(ID,GetPort(pBuff));
		}else if(ID < 8){
			shell_output(s, 1, g_Lcd3, GetLCDPin(ID), GetLCDPinID(ID));
		}
	}

	else if(strcmp_P(g_Buffer,g_AddOP) == 0){
		g_OP = 1;
		g_ID = ID;
	}
	else if(strcmp_P(g_Buffer,g_UpOP) == 0){
		g_OP = 2;
		g_ID = ID;
	}
	else if(strcmp_P(g_Buffer,g_RemOP) == 0){
		g_OP = 3;
		g_ID = ID;
	}
	else if(strcmp_P(g_Buffer,g_DownOP) == 0){
		g_OP = 4;
		g_ID = ID;
	}

	else if(strcmp_P(g_Buffer,g_Input) == 0){
		if(Update){
			SetOptIn(ID, GetPort(pBuff));
		}else{
			shell_output(s, 1, g_Opt, ID, g_Input, GetPortID(GetOptIn(ID)));
		}
	}
	else if(strcmp_P(g_Buffer,g_InVal) == 0){
		switch(GetPortIO(GetOptIn(ID)))
		{
		case OUTPUT:
		case INPUT:
		case MEMORY:
		case MEMORY_SW:
			if(Update){
				SetOptInVal(ID, GetComboVal(pBuff, g_OptsBinV));
			}else{
				shell_output(s, 1, g_Opt, ID, g_InVal, GetComboOpt(GetOptInVal(ID),g_OptsBinV,g_Buffer));
			}
			break;
		case INPUT_ADC:
		case OUTPUT_PWM:
			if(Update){
				SetOptInVal(ID, GetComboVal(pBuff, g_OptsAdcV));
			}else{
				shell_output(s, 1, g_OptN, ID, g_InVal, GetOptInVal(ID));
			}
			break;
		}
	}
	else if(strcmp_P(g_Buffer,g_InAux) == 0){
		switch(GetPortIO(GetOptIn(ID)))
		{
		case OUTPUT:
		case INPUT:
		case MEMORY:
		case MEMORY_SW:
			if(GetOptInVal(ID) < AND)
			{
				if(Update){
					SetOptAux(ID,atoi(pBuff));
				}else{
					shell_output(s, 1, g_OptN, ID, g_InAux, GetOptAux(ID));
				}
			}
			else
			{
				if(Update){
					SetOptAux(ID,GetPort(pBuff));
				}else{
					shell_output(s, 1, g_Opt, ID, g_InAux, GetPortID(GetOptAux(ID)));
				}
			}
			break;
		case INPUT_ADC:
		case OUTPUT_PWM:
			if(GetOptInVal(ID) < EQ_CHN)
			{
				if(Update){
					SetOptAux(ID,atoi(pBuff));
				}else{
					shell_output(s, 1, g_OptN, ID, g_InAux, GetOptAux(ID));
				}
			}
			else
			{
				if(Update){
					SetOptAux(ID,GetPort(pBuff));
				}else{
					shell_output(s, 1, g_Opt, ID, g_InAux, GetPortID(GetOptAux(ID)));
				}
			}
			break;
		}
	}
	else if(strcmp_P(g_Buffer,g_Output) == 0){
		if(Update){
			SetOptOut(ID, GetPort(pBuff));
		}else{
			shell_output(s, 1, g_Opt, ID, g_Output, GetPortID(GetOptOut(ID)));
		}
	}
	else if(strcmp_P(g_Buffer,g_OutVal) == 0){
		switch(GetPortIO(GetOptOut(ID)))
		{
		case MEMORY:
		case MEMORY_SW:
		case OUTPUT:
			if(Update){
				SetOptOutVal(ID,GetComboVal(pBuff, g_OptsBitV));
			}else{
				shell_output(s, 1, g_Opt, ID, g_OutVal, GetComboOpt(GetOptOutVal(ID),g_OptsBitV,g_Buffer));
			}
			break;
		case OUTPUT_PWM:
			if(Update){
				SetOptOutVal(ID,atoi(pBuff));
			}else{
				shell_output(s, 1, g_OptN, ID, g_OutVal, GetOptOutVal(ID));
			}
			break;
		}
	}

	else if(strcmp_P(g_Buffer,g_Value) == 0){
		switch(GetPortIO(ID))
		{
		case INPUT:
		case OUTPUT:
		case MEMORY:
		case MEMORY_SW:
			if(Update){
				if(strcmp_P(pBuff, g_Hi) == 0)
					SetBit(ID,1);
				if(strcmp_P(pBuff, g_Lo) == 0)
					SetBit(ID,0);
			}else{
				shell_output(s, 1, g_Val, g_Value, ID, GetBit(ID) ? g_Hi : g_Lo);
			}
			break;
		case INPUT_ADC:
		case OUTPUT_PWM:
			if(Update){
				SetVal(ID,atoi(pBuff));
			}else{
				shell_output(s, 1, g_ValN, g_Value, ID, GetVal(ID));
			}
			break;
		}
	}

	else if(strcmp_P(g_Buffer,g_AddPG) == 0){
		g_OP = 5;
		g_ID = ID;
	}
	else if(strcmp_P(g_Buffer,g_UpPG) == 0){
		g_OP = 6;
		g_ID = ID;
	}
	else if(strcmp_P(g_Buffer,g_RemPG) == 0){
		g_OP = 7;
		g_ID = ID;
	}
	else if(strcmp_P(g_Buffer,g_DownPG) == 0){
		g_OP = 8;
		g_ID = ID;
	}

	else if(strncmp_P(g_Buffer,g_Page,strlen_P(g_Page)) == 0){
		int PG = (char)g_Buffer[strlen_P(g_Page)]-'A';
		if(Update){
			SetScr(ID,PG,pBuff);
		}else{
			char* scr = GetScr(ID,PG);
			if(scr){
				shell_output(s, 1, g_Lcd1, g_Page, scr);
			}
		}
	}
	else if(strncmp_P(g_Buffer,g_Write,strlen_P(g_Write)) == 0){
		int PG = (char)g_Buffer[strlen_P(g_Write)]-'A';
		if(Update){
			LockIndex();
			set_cursor(0,PG+1);
			lcd_string(pBuff);
		}
	}
	else if(strcmp_P(g_Buffer,g_Action) == 0){
		if(Update){
			SetScrAction(ID,GetPort(pBuff));
		}else{
			shell_output(s, 1, g_Lcd1, g_Action, GetPortID(GetScrAction(ID)));
		}
	}
	else if(strcmp_P(g_Buffer,g_Auto) == 0){
		if(Update){
			SetLCDAuto(atoi(pBuff));
		}else{
			shell_output(s, 1, g_Lcd2, g_Auto, GetLCDAuto());
		}
	}

	else if(strcmp_P(g_Buffer,g_Save) == 0){
		g_OP = 9;
		g_ID = ID;
	}

	else if(strlen(g_Buffer) > 0){
		shell_output(s, 1, g_Unknown, g_Buffer);
	}

	if(s)
		shell_queue();
	shell_output(s, 0, SHELL_PROMPT);
}
